/*
 * Test pipe operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <thread>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <fsclient/client.h>
#include <fsclient/file.h>
#include <mem/memory_utils.h>



/* Minimum pipe region size in pages. */

const unsigned int PIPE_PAGES = 2;

/* Use the writer to fill the pipe with data. */

static void write_pipe(file_t *writer)
{
  offset_t size = 600;
  char buffer[size];

  for (int loop = 0; loop < 3; loop++)
  {
    for (int region = 0; region < 26; region++)
    {
      memset(buffer, (int) 'a' + region, size);

      offset_t nwritten = client_write(writer, buffer, size);

      printf("Written %ld/%ld in #%d of %d/%d to pipe...\n", nwritten, size, region, loop, 2);
    }
  }

  /* Flush to make the final output available. */

  client_flush(writer);
}

/* Use the reader to obtain data from the pipe. */

static void read_pipe(file_t *reader)
{
  offset_t size = 600;
  char buffer[size];
  offset_t nread;

  do
  {
    nread = client_read(reader, buffer, size);

    printf("Read %ld/%ld from pipe...\n", nread, size);

    for (offset_t i = 0; i < nread; i += 60)
    {
      fwrite(buffer + i, sizeof(char), nread - i > 60 ? 60 : nread - i, stdout);
      fputs("\n", stdout);
    }
  }
  while (nread);

  printf("Data shown.\n");
}

int main(void)
{
  /* Obtain access to the filesystem. */

  l4_cap_idx_t server = l4re_env_get_cap("server");

  /* Invoke the open method to receive the file reference. */

  file_t reader, writer;
  long err = pipe_open(page(PIPE_PAGES), &reader, &writer, server);

  if (err)
  {
    printf("Could not obtain pipe: %s\n", l4sys_errtostr(err));
    return 1;
  }

  /* Make the reader and writer blocking to permit synchronisation. */

  if ((err = client_set_blocking(&reader, NOTIFY_CONTENT_AVAILABLE)) ||
      (err = client_set_blocking(&writer, NOTIFY_SPACE_AVAILABLE)))
  {
    printf("Could not set as blocking: %s\n", l4sys_errtostr(err));
    return 1;
  }

  /* Schedule reader and writer threads. */

  std::thread *activities[2];

  activities[0] = new std::thread(read_pipe, &reader);
  activities[1] = new std::thread(write_pipe, &writer);

  for (int i = 0; i < 2; i++)
    activities[i]->join();
}

// vim: tabstop=2 expandtab shiftwidth=2
